"""The tests for the geojson platform."""
import datetime
from asynctest.mock import patch, MagicMock

from homeassistant.components import geo_location
from homeassistant.components.geo_location import ATTR_SOURCE
from homeassistant.components.geo_location.nsw_rural_fire_service_feed import \
    ATTR_EXTERNAL_ID, SCAN_INTERVAL, ATTR_CATEGORY, ATTR_FIRE, ATTR_LOCATION, \
    ATTR_COUNCIL_AREA, ATTR_STATUS, ATTR_TYPE, ATTR_SIZE, \
    ATTR_RESPONSIBLE_AGENCY, ATTR_PUBLICATION_DATE
from homeassistant.const import CONF_URL, EVENT_HOMEASSISTANT_START, \
    CONF_RADIUS, ATTR_LATITUDE, ATTR_LONGITUDE, ATTR_FRIENDLY_NAME, \
    ATTR_UNIT_OF_MEASUREMENT, ATTR_ATTRIBUTION
from homeassistant.setup import async_setup_component
from tests.common import assert_setup_component, async_fire_time_changed
import homeassistant.util.dt as dt_util

URL = 'http://geo.json.local/geo_json_events.json'
CONFIG = {
    geo_location.DOMAIN: [
        {
            'platform': 'nsw_rural_fire_service_feed',
            CONF_URL: URL,
            CONF_RADIUS: 200
        }
    ]
}


def _generate_mock_feed_entry(external_id, title, distance_to_home,
                              coordinates, category=None, location=None,
                              attribution=None, publication_date=None,
                              council_area=None, status=None,
                              entry_type=None, fire=True, size=None,
                              responsible_agency=None):
    """Construct a mock feed entry for testing purposes."""
    feed_entry = MagicMock()
    feed_entry.external_id = external_id
    feed_entry.title = title
    feed_entry.distance_to_home = distance_to_home
    feed_entry.coordinates = coordinates
    feed_entry.category = category
    feed_entry.location = location
    feed_entry.attribution = attribution
    feed_entry.publication_date = publication_date
    feed_entry.council_area = council_area
    feed_entry.status = status
    feed_entry.type = entry_type
    feed_entry.fire = fire
    feed_entry.size = size
    feed_entry.responsible_agency = responsible_agency
    return feed_entry


async def test_setup(hass):
    """Test the general setup of the platform."""
    # Set up some mock feed entries for this test.
    with patch('geojson_client.nsw_rural_fire_service_feed.'
               'NswRuralFireServiceFeed') as mock_feed:
        mock_entry_1 = _generate_mock_feed_entry(
            '1234', 'Title 1', 15.5, (-31.0, 150.0), category='Category 1',
            location='Location 1', attribution='Attribution 1',
            publication_date=datetime.datetime(2018, 9, 22, 8, 0,
                                               tzinfo=datetime.timezone.utc),
            council_area='Council Area 1', status='Status 1',
            entry_type='Type 1', size='Size 1', responsible_agency='Agency 1')
        mock_entry_2 = _generate_mock_feed_entry('2345', 'Title 2', 20.5,
                                                 (-31.1, 150.1),
                                                 fire=False)
        mock_entry_3 = _generate_mock_feed_entry('3456', 'Title 3', 25.5,
                                                 (-31.2, 150.2))
        mock_entry_4 = _generate_mock_feed_entry('4567', 'Title 4', 12.5,
                                                 (-31.3, 150.3))
        mock_feed.return_value.update.return_value = 'OK', [mock_entry_1,
                                                            mock_entry_2,
                                                            mock_entry_3]

        utcnow = dt_util.utcnow()
        # Patching 'utcnow' to gain more control over the timed update.
        with patch('homeassistant.util.dt.utcnow', return_value=utcnow):
            with assert_setup_component(1, geo_location.DOMAIN):
                assert await async_setup_component(
                    hass, geo_location.DOMAIN, CONFIG)
                # Artificially trigger update.
                hass.bus.async_fire(EVENT_HOMEASSISTANT_START)
                # Collect events.
                await hass.async_block_till_done()

                all_states = hass.states.async_all()
                assert len(all_states) == 3

                state = hass.states.get("geo_location.title_1")
                assert state is not None
                assert state.name == "Title 1"
                assert state.attributes == {
                    ATTR_EXTERNAL_ID: "1234", ATTR_LATITUDE: -31.0,
                    ATTR_LONGITUDE: 150.0, ATTR_FRIENDLY_NAME: "Title 1",
                    ATTR_CATEGORY: "Category 1", ATTR_LOCATION: "Location 1",
                    ATTR_ATTRIBUTION: "Attribution 1",
                    ATTR_PUBLICATION_DATE:
                        datetime.datetime(2018, 9, 22, 8, 0,
                                          tzinfo=datetime.timezone.utc),
                    ATTR_FIRE: True,
                    ATTR_COUNCIL_AREA: 'Council Area 1',
                    ATTR_STATUS: 'Status 1', ATTR_TYPE: 'Type 1',
                    ATTR_SIZE: 'Size 1', ATTR_RESPONSIBLE_AGENCY: 'Agency 1',
                    ATTR_UNIT_OF_MEASUREMENT: "km",
                    ATTR_SOURCE: 'nsw_rural_fire_service_feed'}
                assert round(abs(float(state.state)-15.5), 7) == 0

                state = hass.states.get("geo_location.title_2")
                assert state is not None
                assert state.name == "Title 2"
                assert state.attributes == {
                    ATTR_EXTERNAL_ID: "2345", ATTR_LATITUDE: -31.1,
                    ATTR_LONGITUDE: 150.1, ATTR_FRIENDLY_NAME: "Title 2",
                    ATTR_FIRE: False,
                    ATTR_UNIT_OF_MEASUREMENT: "km",
                    ATTR_SOURCE: 'nsw_rural_fire_service_feed'}
                assert round(abs(float(state.state)-20.5), 7) == 0

                state = hass.states.get("geo_location.title_3")
                assert state is not None
                assert state.name == "Title 3"
                assert state.attributes == {
                    ATTR_EXTERNAL_ID: "3456", ATTR_LATITUDE: -31.2,
                    ATTR_LONGITUDE: 150.2, ATTR_FRIENDLY_NAME: "Title 3",
                    ATTR_FIRE: True,
                    ATTR_UNIT_OF_MEASUREMENT: "km",
                    ATTR_SOURCE: 'nsw_rural_fire_service_feed'}
                assert round(abs(float(state.state)-25.5), 7) == 0

                # Simulate an update - one existing, one new entry,
                # one outdated entry
                mock_feed.return_value.update.return_value = 'OK', [
                    mock_entry_1, mock_entry_4, mock_entry_3]
                async_fire_time_changed(hass, utcnow + SCAN_INTERVAL)
                await hass.async_block_till_done()

                all_states = hass.states.async_all()
                assert len(all_states) == 3

                # Simulate an update - empty data, but successful update,
                # so no changes to entities.
                mock_feed.return_value.update.return_value = 'OK_NO_DATA', None
                # mock_restdata.return_value.data = None
                async_fire_time_changed(hass, utcnow +
                                        2 * SCAN_INTERVAL)
                await hass.async_block_till_done()

                all_states = hass.states.async_all()
                assert len(all_states) == 3

                # Simulate an update - empty data, removes all entities
                mock_feed.return_value.update.return_value = 'ERROR', None
                async_fire_time_changed(hass, utcnow +
                                        2 * SCAN_INTERVAL)
                await hass.async_block_till_done()

                all_states = hass.states.async_all()
                assert len(all_states) == 0
